@@ -19,6 +19,7 @@
 #define USE_RINTERNALS
 #include <Rinternals.h>
 #include <R_ext/Mathlib.h>
+#include <Rconfig.h>
 
 #include <hdf5.h>
 
@@ -33,8 +34,10 @@ void warningcall (SEXP call, char *format, ...);
 #define errorcall Rf_errorcall
 void errorcall (SEXP call, char *format, ...);
 
-/* global variable */
-int hdf5_global_verbosity;
+/* global variables */
+int hdf5_global_verbosity = 0;
+int hdf5_global_nametidy = 0;
+int hdf5_global_attrcnt;
 
 #if ((H5_VERS_MAJOR > 1) || (H5_VERS_MAJOR == 1 && H5_VERS_MINOR >= 4))
 #define count_size_t hsize_t
@@ -42,6 +45,23 @@ int hdf5_global_verbosity;
 #define count_size_t size_t
 #endif
 
+void
+nametidy (char *name)
+{
+  /* This function operates on a string to remove any chars regarded
+     by R as not suitable for a variable name i.e. the first char must
+     be alpha or . and the rest can also be digits */
+  unsigned i;
+
+  if (!isalpha (name[0]) && name[0] !='.')
+    name[0] = '.';
+
+  for (i = 1; i < strlen (name); i++)
+    if(!isalnum (name[i]) && name[i] !='.')
+      name[i] = '.';
+}
+
+
 static herr_t
 ref_string (hid_t sid, hid_t did, H5T_cdata_t *cdata,
 	    count_size_t count, size_t stride, size_t bkg_stride,
@@ -78,13 +98,28 @@ string_ref (hid_t sid, hid_t did, H5T_cdata_t *cdata,
     {
       size_t size = H5Tget_size (sid);
       unsigned char srcbuf[size * count], *srcptr = srcbuf;
+      unsigned char cstring[size+1];
       size_t i;
 
+      /* `count' is the number of strings in the array. size is the
+         number of chars in each string.  HDF5 string arrays are not
+         ragged, so there is one value of size for all the elements of
+         the dataset */
+      if (hdf5_global_verbosity > 1)
+	Rprintf ("in string_ref: count=%d, size=%d srcbf=%d\n",
+                 (int) count,(int) size, sizeof(srcbuf));
+      /* Copy the whole array of strings from buf to srcbuf */
       memcpy (srcbuf, buf, sizeof (srcbuf));
-      srcbuf[sizeof (srcbuf) - 1] = '\0';
+      /* now copy the strings into an R string array, one at a time */
       for (i = 0; i < count;i ++)
-	{
-	  ((SEXPREC **)buf)[i] = mkChar (srcptr);
+        {
+	  /* The next 2 lines are necessary to ensure that each
+	     element of the array is null-terminated.  HDF5 files can
+	     contain arrays of Fortran strings which are space-padded,
+	     not null-terminated */
+	  strncpy (cstring, srcptr, size);
+	  cstring[size]='\0';
+	  ((SEXPREC **)buf)[i] = mkChar (cstring);
 	  srcptr += size;
 	}
     }
@@ -101,31 +136,36 @@ struct permute_info {
   hid_t dataset;
   hid_t memtid;
   hid_t space;
-  hid_t mspace;
+  long tmpcnt;
   void *buf;
+  void *tmpbuf;
 };
 
 static void
 permute (struct permute_info *pinfo, unsigned dimnum)
 {
+  /* With large datasets, a great deal too much time (sometimes a
+   * quasi-infinite amount!) was spent in this function. This was
+   * because it was reading the dataset one element at a time. HCP
+   * changed this so that the data is read in one go into tmpbuf
+   * _before_ permute is called. permute then rearranges it.*/
+     
   hssize_t i;
 
   if (dimnum < pinfo->rank)
     {
+
       for (i = 0; i < pinfo->dims[dimnum]; i++)
 	{
 	  pinfo->coord[dimnum] = i;
 	  permute (pinfo, dimnum + 1);
 	}
+
     }
   else
     {
       unsigned offset, mult;
-
-      if (H5Sselect_elements (pinfo->space, H5S_SELECT_SET, 1,
-			      (const hssize_t **) pinfo->coord) < 0)
-	errorcall (pinfo->call, "Unable to select file elements");
-
+      
       offset = pinfo->coord[0];
       mult = 1;
       for (i = 1; i < pinfo->rank; i++)
@@ -134,52 +174,69 @@ permute (struct permute_info *pinfo, unsigned dimnum)
 	  offset += pinfo->coord[i] * mult;
 	}
 
-      {
-	void *pointaddr;
-        
-	switch (pinfo->type)
+      switch (pinfo->type)
+	{
+	case STRSXP: 
+	case VECSXP:
 	  {
-	  case STRSXP: 
-          case VECSXP:
-            pointaddr = &((SEXPREC **)pinfo->buf)[offset];
-            break;
-	  case REALSXP:
-	    pointaddr = &((double *)pinfo->buf)[offset];
-	    break;
-	  case INTSXP: 
-	    pointaddr = &((int *)pinfo->buf)[offset];
-            break;
-          case LGLSXP:
-	    pointaddr = &((int *)pinfo->buf)[offset];
-	    break;
-	  default:
-	    errorcall (pinfo->call, "No support for R type: %d", pinfo->type);
-	    pointaddr = &offset;/* unreached; -Wall */
+	    char **pointaddr, **tmpaddr;
+ 
+	    pointaddr = &((char **) pinfo->buf)[offset];
+	    tmpaddr = &((char **) pinfo->tmpbuf)[pinfo->tmpcnt];
+	    if (pinfo->writeflag)
+	      *tmpaddr = *pointaddr;
+	    else
+	      *pointaddr = *tmpaddr;
 	  }
+	  
+	  break;
+	case REALSXP:
+	  {
+	    double *pointaddr, *tmpaddr;
+	    
+	    pointaddr = &((double *) pinfo->buf)[offset];
+	    tmpaddr = &((double *) pinfo->tmpbuf)[pinfo->tmpcnt];
+	    if (pinfo->writeflag)
+	      *tmpaddr = *pointaddr;
+	    else
+	      *pointaddr = *tmpaddr;
+	  }
+	  break;
+	case INTSXP:
+	  { 
+	    int *pointaddr, *tmpaddr;
 
-	if (pinfo->writeflag)
-	  {
-	    if (H5Dwrite (pinfo->dataset,
-			  pinfo->memtid,
-			  pinfo->mspace,
-			  pinfo->space,
-			  H5P_DEFAULT,
-			  pointaddr) < 0)
-	      errorcall (pinfo->call, "Unable to write dataset");
+	    pointaddr = &((int *) pinfo->buf)[offset];
+	    tmpaddr = &((int * ) pinfo->tmpbuf)[pinfo->tmpcnt];
+	    if (pinfo->writeflag)
+	      *tmpaddr = *pointaddr;
+	    else
+	      *pointaddr = *tmpaddr;
 	  }
-	else
+	  break;
+	case LGLSXP:
 	  {
-	    if (H5Dread (pinfo->dataset,
-			 pinfo->memtid,
-			 pinfo->mspace,
-			 pinfo->space,
-			 H5P_DEFAULT,
-			 pointaddr) < 0)
-	      errorcall (pinfo->call, "Unable to read dataset");
+	    int *pointaddr;
+            unsigned char *tmpaddr;
+
+            pointaddr = &((int *) pinfo->buf)[offset];
+            tmpaddr = &((unsigned char *) pinfo->tmpbuf)[pinfo->tmpcnt];
+
+	    if (pinfo->writeflag)
+                *tmpaddr = *pointaddr;
+            else
+              *pointaddr = *tmpaddr;
 	  }
-      }
+	  break;
+	default:
+	  errorcall (pinfo->call, "No support for R type: %d", pinfo->type);
+	  /*pointaddr = &offset; unreached; -Wall */
+	}
+      pinfo->tmpcnt++;
+      
     }
-}
+  return ;
+} /* end of permute */
 
 static hid_t
 make_sexp_ref_type (SEXP call)
@@ -286,41 +343,110 @@ vector_io (SEXP call, int writeflag, hid_t dataset, hid_t space, SEXP obj)
   int rank = H5Sget_simple_extent_ndims (space);
   hsize_t mdims[1] = {1};
   hsize_t dims[rank], maxdims[rank];
+  /* WARNING: Don't make bufsize a hsize_t. If you do this,
+     then bufsize=-1 ; if(bufsize>0) is (ludicrously) regarded as
+     true. We therefore make bufsize a long and cast it to a hsize_t
+     when we pass it to a HDF5 routine. There are known weirdnesses
+     regarding hsize_t and gcc which may be a compiler bug and this is
+     probably another example.  */
+  long bufsize;
   SEXPTYPE type = TYPEOF (obj);
-  hid_t memtid, tid, mspace;
+  hid_t memtid, tid, plist;
   void *buf;
-
+  void *tmpbuf;
+  long n_elements,i;
+  if (hdf5_global_verbosity > 3)
+    Rprintf ("in vector_io: rank=%d\n",rank); 
   if ((tid = H5Dget_type (dataset)) < 0)
     errorcall (call, "Unable to get type for dataset");
+  
+  if (H5Sget_simple_extent_dims (space, dims, maxdims) < 0)
+    errorcall (call, "Unable to get dimensions of space");
+
+  /* calculate the total number of elements in the dataset */
+  n_elements = 1;
+  for (i = 0; i < rank; i++)
+    {
+      if (hdf5_global_verbosity > 3)
+        Rprintf ("in vector_io:size %d = %d into n_elements..",i,dims[i]); 
+      n_elements = n_elements * dims[i];
+      if (hdf5_global_verbosity > 3)
+        Rprintf ("....=%d\n ",n_elements);
+    }
 
   switch (type) {
   case STRSXP:
-    memtid = make_sexp_ref_type (call);
-    buf = STRING_PTR (obj);
+    { 
+      char *charp;
+      memtid = make_sexp_ref_type (call);
+      buf = STRING_PTR (obj);
+      tmpbuf = R_alloc (n_elements, sizeof( charp));
+      /*bufsize=n_elements*sizeof(charp)*10;*/
+      bufsize = -1;
+      /* I can't work out what bufsize should be for strings. Because they
+	 are less likely to come as a multi-Mbyte array we set it to 
+	 -1 and (later) use the default value for the property list.*/
+    }
     break;
   case REALSXP:
     memtid = H5T_NATIVE_DOUBLE;
     buf = REAL (obj);
+    tmpbuf = R_alloc (n_elements, sizeof(double));
+    bufsize = n_elements * sizeof(double);
     break;
   case INTSXP:
     memtid = H5T_NATIVE_INT;
     buf = INTEGER (obj);
+    tmpbuf = R_alloc (n_elements, sizeof (int));
+    bufsize = n_elements * sizeof (int);
     break;
   case LGLSXP:
     memtid = make_boolean_type (call);
     buf = LOGICAL (obj);
+    tmpbuf = R_alloc (n_elements, sizeof (unsigned char));
+    bufsize = n_elements * sizeof (unsigned char);
     break;
   default:
     errorcall (call, "Can't get type for R type: %d (IO)", type);
     /* unreached (-Wall): */ memtid = tid;  buf = &tid;
   }
+
+  /* set type conversion buffer size in property list . In most circs, 
+     you don't have to do this, but if you have an array with dims
+     a x b x c x d x .... and b x c x d x ..... adds up to > 1MB
+     then HDF5 will barf. */
+
+  if (bufsize > 0)
+    {
+      if (hdf5_global_verbosity > 2)
+        Rprintf ("Setting buffer size in plist\n");
+      plist = H5Pcreate(H5P_DATASET_XFER);
+
+    if (H5Pset_buffer (plist, ((hsize_t) bufsize), NULL, NULL) < 0)
+      errorcall (call,"Unable to set buffer size in property list");
+    }
+  else
+    {
+      if (hdf5_global_verbosity > 2)
+        Rprintf ("Using default transfer plist\n");
+      plist = H5P_DEFAULT;
+    }
+  
+  /* Read data into temporary buffer before permuting it */
+  if (!writeflag)
+    {
+      if (hdf5_global_verbosity > 2)
+        Rprintf ("About to read with bufsize = %d\n",bufsize);
+      
+      if (H5Dread (dataset, memtid, H5S_ALL, H5S_ALL, plist, tmpbuf) < 0)
+        errorcall (call, "Unable to read dataset");
+      if (hdf5_global_verbosity > 2)
+        Rprintf (" Done read\n");
+    }
+  
+  if (hdf5_global_verbosity > 2)
+    Rprintf ("in vector_io: permuting\n");  
   
-  if (H5Sget_simple_extent_dims (space, dims, maxdims) < 0)
-    errorcall (call, "Unable to get dimensions of space");
-
-  if ((mspace = H5Screate_simple (1, mdims, NULL)) < 0)
-    errorcall (call, "Unable to create point space");
-
   {
     struct permute_info pinfo;
     hssize_t coord[rank];
@@ -334,16 +460,43 @@ vector_io (SEXP call, int writeflag, hid_t dataset, hid_t space, SEXP obj)
     pinfo.dataset = dataset;
     pinfo.memtid = memtid;
     pinfo.space = space;
-    pinfo.mspace = mspace;
     pinfo.buf = buf;
+    pinfo.tmpbuf=tmpbuf;
+    pinfo.tmpcnt=0;
+
+    /* The grinding slowness happened inside this call to permute()
+       and was caused quite specifically by calling H5Dread() with a
+       dataset size of 1 for every element of the dataset. The fix I
+       have applied is to have an o/p buffer tmpbuf as big as obj that
+       we do the read/write to/from and use permute to transfer data
+       from tmpbuf to buf. If we read directly into buf the
+       indices get scrambled. Whatever permute does to them is
+       necessary.*/
 
     permute (&pinfo, 0);
+
+    
+    if(writeflag)
+      {
+        if (hdf5_global_verbosity > 2)
+          Rprintf ("About to write\n");
+        if (H5Dwrite (dataset, memtid, H5S_ALL, H5S_ALL, plist, tmpbuf) < 0)
+          errorcall (call, "Unable to write dataset");
+        if (hdf5_global_verbosity > 2)
+          Rprintf ("About to write\n");
+      }
   }
+  
+  if (hdf5_global_verbosity > 2)
+    Rprintf ("in vector_io: tidying\n");  
+  if (bufsize > 0)
+    {
+      if (H5Pclose(plist) < 0)	
+        errorcall (call, "Unable to close plist");
+    }
 
-  if (H5Sclose (mspace) < 0)
-    errorcall (call, "Unable to close point space");
-
-  if (type == STRSXP || type == LGLSXP) {
+  if (type == STRSXP || type == LGLSXP)
+    {
       if (H5Tclose (memtid) < 0)
 	errorcall (call, "Unable to close string reference type");
     }
@@ -898,14 +1051,23 @@ static void
 add_to_list (struct hdf5_iterate_info *iinfo, const char *name, SEXP obj)
 {
   SEXP pair;
-  
+  char newname[strlen(name)+1];
+
+  strcpy (newname,name);
+  if (hdf5_global_nametidy)
+    {
+      if (hdf5_global_verbosity > 1)
+        Rprintf(" Tidying name %s ", newname);
+      nametidy (newname);
+    }
   PROTECT (pair = CONS (obj, CDR (iinfo->ret)));
-  TAG (pair) = install ((char *) name);
+  TAG (pair) = install ((char *) newname);
   SETCDR (iinfo->ret, pair);
   UNPROTECT (1);
-
+  if (hdf5_global_verbosity > 1 && hdf5_global_nametidy) 
+    Rprintf (".. to %s \n ",newname);
   if (hdf5_global_verbosity > 2)
-    Rprintf ("Adding `%s' to list\n", name);
+    Rprintf ("Adding `%s' to list\n", newname);
 }
 
 
@@ -1037,16 +1199,19 @@ struct hdf5_attribute_info {
   const char *name;
 };
 
-static herr_t
+/*static*/ 
+herr_t
 hdf5_process_attribute (hid_t loc_id, const char *attrName, void *data)
 {
   struct hdf5_attribute_info *ainfo = data;
   hid_t aid, sid, tid;
   H5T_class_t class;
   size_t tid_size;
-
+  char newname[strlen(attrName +1)];
+  hdf5_global_attrcnt++;
   if (hdf5_global_verbosity > 1)
-    Rprintf ("Processing attribute %s\n",attrName);
+    Rprintf ("Processing attribute %d called %s\n",
+	     hdf5_global_attrcnt,attrName);
 
   if ((aid = H5Aopen_name (loc_id, attrName)) < 0)
     errorcall (ainfo->call, "could not open attribute `%s'", attrName);
@@ -1067,24 +1232,39 @@ hdf5_process_attribute (hid_t loc_id, const char *attrName, void *data)
 	       attrName);
 
   {
-    int rank;
-
+    int rank,r_rank;
+    
     if ((rank = H5Sget_simple_extent_ndims (sid)) < 0)
       errorcall (ainfo->call,
 		 "could not get rank of attribute space `%s'",
 		 attrName);
-
+    
     if (hdf5_global_verbosity > 1) 
       Rprintf ("attribute %s has rank %d \n",attrName,rank);    
+
+    if (rank == 0)
+      r_rank = 1;
+    else
+      r_rank = rank;
+
     {
-      hsize_t dims[rank];
+      hsize_t dims[r_rank];
 
-      if (H5Sget_simple_extent_dims (sid, dims, NULL) < 0)
-	errorcall (ainfo->call,
-		   "could not get extent of attribute space `%s'",
-		   attrName);
-
-      if (rank == 1)
+      if(rank == 1)
+        {
+          if (H5Sget_simple_extent_dims (sid, dims, NULL) < 0)
+            errorcall (ainfo->call,
+                       "could not get extent of attribute space `%s'",
+                       attrName);
+        }
+      else
+        {
+          /* scalar, but we'll pretend it is a vector of length 1 */
+          dims[0] = 1; 
+          if (hdf5_global_verbosity > 2) 
+            Rprintf ("Rank 0 attribute treated as rank 1 size 1\n");
+        }
+      if (r_rank == 1)
 	{
 	  unsigned count = dims[0];
 	  SEXPTYPE type;
@@ -1112,6 +1292,8 @@ hdf5_process_attribute (hid_t loc_id, const char *attrName, void *data)
 	      break;
 	    case H5T_STRING:
 	      type = STRSXP;
+	      if (hdf5_global_verbosity > 2) 
+		Rprintf ("Attribute is a string\n");
 	      memtid = make_sexp_ref_type (ainfo->call);
 	      break;
 	    default:
@@ -1139,10 +1321,21 @@ hdf5_process_attribute (hid_t loc_id, const char *attrName, void *data)
 	    }
 	  if (H5Aread (aid, memtid, buf) < 0)
 	    errorcall (ainfo->call, "unable to read attribute `%s'", attrName);
-
+	  if (hdf5_global_verbosity > 2) 
+	    Rprintf ("string length of new name =%d\n",strlen(attrName+1));
+	  strcpy(newname,attrName);
+	  if (hdf5_global_nametidy)
+            {
+              if (hdf5_global_verbosity > 1) 
+                Rprintf (" Tidying attribute name %s ",newname);
+              nametidy (newname);
+              if (hdf5_global_verbosity > 1) 
+                Rprintf("....to %s\n",newname);
+            }
+	  
 	  if (!isNull (ainfo->obj))
-	    setAttrib (ainfo->obj, install ((char *) attrName), vec);
-
+	    setAttrib (ainfo->obj, install ((char *) newname), vec);
+	  
 	  UNPROTECT (1);
 
 	  if (H5Tclose (memtid) < 0)
@@ -1164,20 +1357,29 @@ hdf5_process_attribute (hid_t loc_id, const char *attrName, void *data)
     errorcall (ainfo->call, "unable to close attribute `%s'", attrName);
   if (hdf5_global_verbosity > 1) 
     Rprintf ("Done processing attribute %s\n",attrName);
-  return 0;
+  if (hdf5_global_attrcnt > 100)
+    {
+      Rprintf ("WTF? More than 100 attributes? \n");
+      return ((herr_t) 99);
+      /* and here we sometimes get a seggie. Why?*/
+    }
+  else
+    return 0;
 }
 
 static void
 hdf5_load_attributes (SEXP call, hid_t id, SEXP obj, const char *name)
 {
-  unsigned idx = 0;
+  unsigned *idx = NULL;
   struct hdf5_attribute_info ainfo;
+  herr_t retval;
 
   ainfo.call = call;
   ainfo.obj = obj;
   ainfo.name = name;
-
-  if (H5Aiterate (id, &idx, hdf5_process_attribute, &ainfo) < 0)
+  hdf5_global_attrcnt = 0;
+  retval = H5Aiterate (id, idx, hdf5_process_attribute, &ainfo);
+  if (retval < 0)
     errorcall (call, "unable to iterate over attributes");
 }
 
@@ -1436,12 +1638,25 @@ hdf5_process_object (hid_t id, const char *name, void *client_data)
                     }
                   if (hdf5_global_verbosity > 2)
                     Rprintf ("\n");
-                  PROTECT (obj = allocArray (type, dimsvector)); 
+                  if (hdf5_global_verbosity > 2)
+                    Rprintf ("about to do actual allocation\n");
+                  PROTECT (obj = allocArray (type, dimsvector));
+                  if (hdf5_global_verbosity > 2)
+                    Rprintf ("done allocation -- unprotecting\n");
                   UNPROTECT (2);
+                  if (hdf5_global_verbosity > 2)
+                    Rprintf ("done unprotecting -- re-protecting obj\n"); 
                   PROTECT (obj);
+		  if (hdf5_global_verbosity > 2)
+		    Rprintf ("done re-protecting\n");  
+
                 }
+	      if (hdf5_global_verbosity > 2)
+		Rprintf ("calling vector_io. Hangs here with big datsets\n");  
               vector_io (iinfo->call, LFALSE, dataset, space, obj);
-              iinfo->add (iinfo, name, obj);
+	      if (hdf5_global_verbosity > 2)
+		Rprintf ("Phew. Done it. calling iinfo->add\n");  
+	      iinfo->add (iinfo, name, obj);
               if (hdf5_global_verbosity > 2)
                 {
                   Rprintf ("Rank > 1 or not VECSXP\n");
@@ -1473,7 +1688,17 @@ add_to_symbol_table (struct hdf5_iterate_info *iinfo,
 		     const char *name,
 		     SEXP obj)
 {
-  setVar (install ((char *)name), obj, iinfo->env);
+  char newname[strlen(name)+1];
+  strcpy(newname,name);
+  if(hdf5_global_nametidy)
+    {
+      if (hdf5_global_verbosity > 1)
+        Rprintf("Tidying name %s ",newname);
+      nametidy (newname);
+    }
+  setVar (install ((char *)newname), obj, iinfo->env);
+  if (hdf5_global_verbosity > 1 && hdf5_global_nametidy)
+    Rprintf (".. to %s \n ",newname);
 }
 
 SEXP do_hdf5load (SEXP args)
@@ -1494,14 +1719,14 @@ SEXP do_hdf5load (SEXP args)
   if (TYPEOF (CADR (args)) != LGLSXP)
     errorcall (call, "second argument must be a logical vector");
 
-  path = CHAR(STRING_ELT (CAR(args), 0));
+  path = CHAR (STRING_ELT (CAR(args), 0));
   restore_syms = INTEGER (CADR(args))[0];
 
-  hdf5_global_verbosity = INTEGER (CADDR(args))[0];
+  hdf5_global_verbosity = INTEGER (CADDR (args))[0];
   if (hdf5_global_verbosity > 2) 
     Rprintf ("hdf5_global_verbosity=%d load=%d\n",
              hdf5_global_verbosity, restore_syms);
-
+  hdf5_global_nametidy  = INTEGER (CADDDR (args))[0];
   H5dont_atexit ();
   
   if ((fid = H5Fopen (path, H5F_ACC_RDONLY, H5P_DEFAULT)) < 0)
